// addmix
// combines an arbitrary number of input files together into a single output file, with optional input level control.
// CPU demands: low.
// Memory requirements: 2x size of longest input file.

#![allow(unused_parens)]

use std::io::prelude::*;
use std::fs::File;
use std::env::args;
use std::process::exit;

fn main()
{
	let argv = args().collect::<Vec<String>>();
	if argv.len() < 3
	{
		eprintln!("usage: addmix <output.dat> <input_1.dat>:<level> <input_2.dat>:<level> ...");
		return;
	}

	let output_filename = argv[1].clone();
	let mut output_file = match File::create(&argv[1])
	{
		Err(reason) => 
		{
			eprintln!("addmix: could not open/create specified output file {} for writing: {}", argv[1], reason);
			exit(1);
		},
		Ok(file) => file,
	};
	let mut input_files:Vec<(File,String,f32)> = Vec::new();
	for i in 2..argv.len()
	{
		let mut new_filename = argv[i].clone();
		let mut new_level:f32 = 1.0;
		if let Some((filename,levelstring)) = argv[i].rsplit_once(':')
		{
			if let Ok(val) = levelstring.parse::<f32>()
			{
				// decide how to handle the filename:level vs just filename syntax by testing to see if what's after the colon
				// actually looks like a number. if it doesn't, silently assume that the whole thing was supposed to be a colon-containing filename.
				// if this wasn't the case, and the user meant to type a number, then they should be able to understand what happened from the
				// "file not found" error message.
				new_level = val;
				new_filename = filename.to_string();
			}
		}
		let new_file = match File::open(&new_filename)
		{
			Err(reason) =>
			{
				eprintln!("addmix: could not open specified input file {} for reading: {}", new_filename, reason);
				exit(1);
			},
			Ok(file) => file,
		};
		input_files.push((new_file, new_filename, new_level));
	}

	let mut output_samples:Vec<f32> = Vec::new();

	for (file, filename, level) in input_files.iter_mut()
	{
		let mut file_data:Vec<u8> = Vec::new();
		match file.read_to_end(&mut file_data)
		{
			Err(reason) => 
			{
				eprintln!("addmix: could not read from input file {}: {}", filename, reason);
				exit(1);
			},
			Ok(_) => (),
		};

		let file_length:usize = file_data.len()/4;
		for _ in 0..file_length.saturating_sub(output_samples.len())
		{
			output_samples.push(0.0);
		}

		for (t,chunk) in file_data.chunks_exact(4).enumerate()
		{
			let mut sample_bytes:[u8;4] = [0x00; 4];
			sample_bytes.copy_from_slice(chunk);

			output_samples[t] += *level * f32::from_be_bytes(sample_bytes);
		}
	}

	let mut output_data:Vec<u8> = Vec::with_capacity(output_samples.len() * 4);
	for sample in output_samples.iter()
	{
		output_data.extend_from_slice(&sample.to_be_bytes());
	}
	match output_file.write_all(&output_data)
	{
		Err(reason) => 
		{
			eprintln!("addmix: could not write to output file {}: {}", output_filename, reason);
			exit(1);
		},
		Ok(_) => (),
	};
}
