#![allow(unused_parens)]

/*
echospace - rudimentary 3D echo simulator

accepts a signal, a transmitter position, a receiver position, and a scene matrix,
and emits a simulated received signal

! limitations !
- assumes transmitter and receiver to be isotropically radiant/sensitive
- does not simulate:
	-- occlusion                        (hard)
	-- secondary reflections            (computationally intensive)
	-- any frequency-dependent effects  (needs frequency-domain operations, currently not implemented)
	-- any nonlinear effects            (significance unknown)
*/

use std::io::prelude::*;
use std::fs::File;
use std::env::args;

fn distance
(
	a: &[f32;3],
	b: &[f32;3],
) ->
	f32
{
	(
		(a[0] - b[0]).powi(2)
		+
		(a[1] - b[1]).powi(2)
		+
		(a[2] - b[2]).powi(2)
	).sqrt()
}

fn skew
(
	input: &[f32],
	output: &mut [f32],
	offset: f32,
) {
	let skew_factor:f32 = offset.fract();
	let output_slice:&mut [f32] = &mut output[0..input.len()+1];
	for (i_point,o_point) in (input.iter()).zip(output_slice.iter_mut())
	{
		*o_point += i_point * (1.0 - skew_factor);
		*o_point += i_point * skew_factor;
	}
}

fn main()
{
	let argv = args().collect::<Vec<String>>();
	if argv.len() != 7
	{
		eprintln!("usage: echospace <txx>,<txy>,<txz> <rxx>,<rxy>,<rxz> <wave_speed> <input_signal.dat> <scene.3da> <output_signal.dat>");
		return;
	}

	// argument             type           description
	// <txx>,<txy>,<txz>    f32,f32,f32    coordinates of the transmitter relative to the scene volume
	// <rxx>,<rxy>,<rxz>    f32,f32,f32    coordinates of the receiver relative to the scene volume
	// <wavespeed>          f32            speed of the wave in the medium, in points (3D scene coordinates) per signal sample
	// <input_signal.dat>   dat file       signal emitted by the transmitter
	// <scene.3da>          3da file       3D array file defining a reflectivity value at every point in the simulated volume (see 3da file definition)
	// <output_signal.dat>  dat file       simulated signal detected by the receiver

	let txpos_string:String = argv[1].clone();
	let rxpos_string:String = argv[2].clone();
	let wave_speed:f32 = argv[3].parse::<f32>().unwrap();
	let mut input_file = File::open(&argv[4]).unwrap();
	let mut scene_file = File::open(&argv[5]).unwrap();
	let mut output_file = File::create(&argv[6]).unwrap();

	let mut txpos:[f32;3] = [0.0, 0.0, 0.0];
	let txpos_comps:Vec<&str> = txpos_string.split(',').collect();
	for n in 0..3
	{
		txpos[n] = txpos_comps[n].parse::<f32>().unwrap();
	}

	let mut rxpos:[f32;3] = [0.0, 0.0, 0.0];
	let rxpos_comps:Vec<&str> = rxpos_string.split(',').collect();
	for n in 0..3
	{
		rxpos[n] = rxpos_comps[n].parse::<f32>().unwrap();
	}
	
	let mut input_data:Vec<u8> = Vec::new();
	input_file.read_to_end(&mut input_data).unwrap();
	let mut input_signal:Vec<f32> = Vec::with_capacity(input_data.len()/4);
	for chunk in input_data.chunks_exact(4)
	{
		let mut sample_bytes:[u8;4] = [0x00; 4];
		sample_bytes.copy_from_slice(chunk);
		input_signal.push(f32::from_be_bytes(sample_bytes));
	}

	let mut scene_xsize_bytes:[u8;4] = [0x00; 4];
	let mut scene_ysize_bytes:[u8;4] = [0x00; 4];
	let mut scene_zsize_bytes:[u8;4] = [0x00; 4];
	let mut scene_data:Vec<u8> = Vec::new();
	scene_file.read_exact(&mut scene_xsize_bytes).unwrap();
	scene_file.read_exact(&mut scene_ysize_bytes).unwrap();
	scene_file.read_exact(&mut scene_zsize_bytes).unwrap();
	scene_file.read_to_end(&mut scene_data).unwrap();
	let scene_xsize:usize = u32::from_be_bytes(scene_xsize_bytes) as usize;
	let scene_ysize:usize = u32::from_be_bytes(scene_ysize_bytes) as usize;
	let scene_zsize:usize = u32::from_be_bytes(scene_zsize_bytes) as usize;
	let mut scene:Vec<f32> = Vec::with_capacity(scene_data.len()/4);
	for chunk in scene_data.chunks_exact(4)
	{
		let mut point_bytes:[u8;4] = [0x00; 4];
		point_bytes.copy_from_slice(chunk);
		scene.push(f32::from_be_bytes(point_bytes));
	}
	// scene[x][y][z] => scene[z*scene_ysize*scene_xsize + y*scene_xsize + x]
	// scene[n] => scene[n % scene_xsize][(n/scene_xsize) % scene_ysize][n/(scene_xsize*scene_ysize)]

	let scene_corners:[[f32;3];8] = 
	[
		[0.0               , 0.0               , 0.0               ],
		[scene_xsize as f32, 0.0               , 0.0               ],
		[0.0               , scene_ysize as f32, 0.0               ],
		[scene_xsize as f32, scene_ysize as f32, 0.0               ],
		[0.0               , 0.0               , scene_zsize as f32],
		[scene_xsize as f32, 0.0               , scene_zsize as f32],
		[0.0               , scene_ysize as f32, scene_zsize as f32],
		[scene_xsize as f32, scene_ysize as f32, scene_zsize as f32],
	];
	let mut max_flight_dist:f32 = 0.0;
	for corner in scene_corners.iter()
	{
		max_flight_dist = max_flight_dist.max(distance(corner,&txpos) + distance(corner,&rxpos));
	}
	let max_flight_time = max_flight_dist*wave_speed;
	eprintln!("maximum simulated path length: {} points", max_flight_dist.ceil());
	eprintln!("maximum simulated path time: {} samples", max_flight_time.ceil());
	
	let mut output_signal:Vec<f32> = vec![0.0; (max_flight_time.floor() as usize) + input_signal.len()+1];

	let mut skewed_input:Vec<f32> = vec![0.0; input_signal.len()+1];

	let tx_rx_flight_dist:f32 = distance(&txpos, &rxpos);
	let tx_rx_flight_time:f32 = tx_rx_flight_dist*wave_speed;
	let tx_rx_flight_samples:usize = tx_rx_flight_time.floor() as usize;
	skew(&input_signal, &mut skewed_input, tx_rx_flight_time.fract());
	for (i_point, o_point) in 
	(
		skewed_input.iter()
	).zip(
		output_signal
		[
			tx_rx_flight_samples
			..
			tx_rx_flight_samples + skewed_input.len()
		].iter_mut()
	) {
		*o_point += i_point;
	}
	
	for x in 0..scene_xsize {
	for y in 0..scene_ysize {
	for z in 0..scene_zsize {
		let pointpos:[f32;3] = [x as f32, y as f32, z as f32];
		let flight_dist:f32 = distance(&pointpos, &txpos) + distance(&pointpos, &rxpos);
		let flight_time:f32 = flight_dist*wave_speed;
		let attenuation:f32 = tx_rx_flight_dist.powi(2) / flight_dist.powi(2);
		let sample_skew:f32 = flight_time.fract();
		let flight_samples:usize = flight_time.floor() as usize;
		let reflectivity:f32 = scene[z*scene_ysize*scene_xsize + y*scene_xsize + x];

		skew(&input_signal, &mut skewed_input, sample_skew);

		for (i_point, o_point) in
		(
			skewed_input.iter()
		).zip(
			output_signal
			[
				flight_samples
				..
				flight_samples + skewed_input.len()
			].iter_mut()
		) {
			*o_point += reflectivity * attenuation * i_point;
		}
	}}}

	let mut output_data:Vec<u8> = Vec::with_capacity(output_signal.len()*4);
	for point in output_signal.iter()
	{
		output_data.extend_from_slice(&point.to_be_bytes());
	}
	
	// write the output file
	output_file.write_all(&output_data).unwrap();
}
